/*******************************************************************
 *
 *	File:		CArlo.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *				Based on code from Otfried Cheong
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

/*
 * Thanks to Ian Jeffray for the following info...

	I'm not sure if the following info is "public knowledge" but I
	know the following virtual addresses in ER5, which might help:

	Page Directory starts at 41000000
	Page Table Info array    41080000  
	Page Tables are at       42000000
	ROM is at                50000000
	Hardware mmap'd IO at    58000000
	Superpage / CPU page     40000000
	RAM Drive at             61000000

	the virt to phys map takes the top 12bits of the virtual address
	as an index in to the 16k page directory table (4096 words).

	The 32bit value which is selected is the PDE (Page Directory Entry).

	The PDE is a pointer to the page table : a 1K table of 32bit items.
	(Ie: One PDE maps 1MB of RAM)

	The next 8bits of the virtual address are an index in to the page
	table, giving the 32bit value which is the start of the physical page.
	(each page is 4K of course).

	The last 12bits of the virtual address select the byte in the
	physical page.


	Hope this isn't teaching folks to suck eggs, and may help?
*/

/*
 * Peter's comments...

	
	Initially I thought that I could find page table n at linear address:
	0x42000000 + n*1024. However that crashed machine for some/most linear
	addresses.
	After some tinkering it seems that page table n can be found by the following:
	Take the n-th TUint32 from base 0x41080000 (Page Table Info array).
	The 16 most significant bit provide the actual index in the Page Tables array.

*/

#include <e32std.h>
#include <e32cons.h>
#include <f32file.h>
#include <e32svr.h>

#include "CArlo.h"
#include "RArlo.h"
#include "CArloConfig.h"


//#define DEBUG_PRINTS_PLEASE

#ifdef DEBUG_PRINTS_PLEASE
extern CConsoleBase*	sConsole;
#define DBGPRINT(x) sConsole->Printf x
#define DBGPRESSKEY(x) sConsole->Printf x;sConsole->Getch()
#else
#define DBGPRINT(x) /**/
#define DBGPRESSKEY(x) /**/
#endif




CArloBase::CArloBase(RArLo& aArlo) : iArlo(aArlo)
{
}


CArloBase::~CArloBase()
{
	delete iTable;
	delete iBootString;
}


void CArloBase::AddPagesL(TInt aBlocks, TUint32 aLoadAddr)
{
	SetupArloPageTableL();

	if (iArloPageTableIndex + aBlocks > NR_PAGE_TABLE_ENTRIES) 
	{
		DBGPRINT((_L("Does not fit in page table.\n")));
		User::Leave(KErrGeneral);
	}
	// add entries to page map 
	for (TInt i = 0; i < aBlocks; i++)
	{
		iArloPageTable[iArloPageTableIndex++] = PlatLinear2Physical(aLoadAddr + i * PAGE_SIZE);
	}
	iArloPageTable[iArloPageTableIndex++] = 0x0;
}

void CArloBase::ArloBoot() 
{
	SetupArloPageTableL();

	TUint32 arloPageTable = PlatLinear2Physical((TUint32)iArloPageTable);
	PlatBoot(arloPageTable, iArloPageTable[2]);
}



TUint CArloBase::GetProcId()
{
	TUint	p;
	iArlo.GetProcessorId(&p);
	return p;
}


void CArloBase::ConstructL()
{
}


void CArloBase::SetBootString(TAny* aBootString, TInt aLength)
{
	SetupArloPageTableL();
	Mem::Copy(iBootString, aBootString, aLength);
	iBootString[aLength] = 0;
}


void CArloBase::SetupArloPageTableL()
{
	if (!iHasSetupArloPageTable)
	{
		iHasSetupArloPageTable = ETrue;

		TText8*		iTable = (TText8*) User::AllocL(PAGE_SIZE * sizeof(TUint32) * 2);
		Mem::FillZ(iTable, PAGE_SIZE * sizeof(TUint32) * 2);
		TText8* table = iTable + (PAGE_SIZE -1);
		// transfer ownership
		iArloPageTable = (TUint32*) (((TUint32) table) & ~(PAGE_SIZE - 1));

		DBGPRINT((_L("iArloPageTable=0x%x\n"), iArloPageTable));

		iBootString = (TText8 *) new (ELeave) TUint32[BOOT_STRING_SIZE]; // word align 
		Mem::FillZ(iBootString, BOOT_STRING_SIZE * sizeof(TUint32));

		// first entry in page map:  phys address of option string
		TUint32 result = PlatLinear2Physical((TUint32) iBootString);

		DBGPRESSKEY((_L("Press a key\n")));

		iArloPageTable[0] = result;

		DBGPRINT((_L("Option buffer at phys address %08x\n"), iArloPageTable[0]));
		DBGPRESSKEY((_L("Press a key\n")));

		// second entry: phys address of memory description
		TUint32		adr = 0x40000030;
		adr = ArloExamine(adr);
		iArloPageTable[1] = PlatLinear2Physical(adr);
		DBGPRINT((_L("Memory description at address %08x\n"), iArloPageTable[1]));
		
		iArloPageTableIndex = 2;
	}
}


#define KMachineSeries7		_L("SERIES 7")
#define KMachineSeries5mx	_L("SERIES5mx")
#define KMachineSeries5		_L("SERIES5 R1")
#define KMachineWins		_L("WINS Pc")

CArloBase* ArloFactory::CreateArloLC(RArLo& aArlo, const TDesC& aMachineName)
{
	CArloBase*		arlo;

#ifdef __WINS__
	aMachineName;		// keep compiler happy
	arlo = new(ELeave)CArloWins(aArlo); 
	arlo->iMachineName = KMachineWins;
#else
	if (aMachineName.CompareF(KMachineSeries5mx) == 0)
	{
		arlo = new(ELeave)CArloSeries5Mx(aArlo);
		arlo->iMachineName = KMachineSeries5mx;
	}
	else if (aMachineName.CompareF(KMachineSeries5) == 0)
	{
		arlo = new(ELeave)CArloSeries5(aArlo); 
		arlo->iMachineName = KMachineSeries5;
	}
	else if (aMachineName.CompareF(KMachineSeries7) == 0)
	{
		arlo = new(ELeave)CArloSeries7(aArlo); 
		arlo->iMachineName = KMachineSeries7;
	}
	else
	{
		arlo = new(ELeave)CArloSeries5Mx(aArlo);
		arlo->iMachineName = KMachineSeries5mx;
	}
#endif
	CleanupStack::PushL(arlo);
	arlo->ConstructL();

	return arlo;
}


/********************************
 *
 * ARM4 *Series5mx & Series 7
 *
 ********************************/

TUint32	CArloArm4::PlatLinear2Physical(TUint32 aLinearAdr)
{
	TUint		pageDirectoryIndex = (aLinearAdr >> 20);
	TUint		pageTableIndex = ((aLinearAdr & 0x000ff000) >> 12);
	TUint		offsetInPage = aLinearAdr & 0xfff;

	DBGPRINT((_L("Virtual adr          =0x%08x\n"), aLinearAdr));
	DBGPRINT((_L("  pageDirectoryIndex =0x%08x\n"), pageDirectoryIndex));
	DBGPRINT((_L("  pageTableIndex     =0x%08x\n"), pageTableIndex));
	DBGPRINT((_L("  offsetInPage       =0x%08x\n"), offsetInPage));


	TUint32 aLevelOne = 0x41000000 + (pageDirectoryIndex << 2);
	DBGPRINT((_L("Level One address %08x\n"), aLevelOne));
	
	aLevelOne = ArloExamine(aLevelOne);
	DBGPRINT((_L("Level One descriptor %08x\n"), aLevelOne));
	
	// we now have the level one descriptor
	if ((aLevelOne & 0x3) == 0x2) 
	{
		DBGPRINT((_L("Section\n")));
		// a Section descriptor
		return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
	}
	if ((aLevelOne & 0x3) != 0x1) 
	{
		DBGPRINT((_L("fault or reserved\n")));
		// Fault or reserved
		return 0;
	}


	TUint32	pageId = 0x41080000 + (pageDirectoryIndex << 2);
	pageId = ArloExamine(pageId);
	DBGPRINT((_L("  page id=0x%08x\n"), pageId));

	TUint32		page = 0x42000000 + ((pageId >> 16) * 1024) + (pageTableIndex << 2);
//	TUint32		page = 0x42000000 + ((pageDirectoryIndex & 0xffff) * 1024) + (pageTableIndex << 2);
	page = ArloExamine(page);
	DBGPRINT((_L("  page address=0x%08x\n"), page));

	TUint32		physAddr = (page & 0xfffff000) +  offsetInPage;
	DBGPRINT((_L("physical address=0x%08x\n"), physAddr));

	DBGPRESSKEY((_L("PlatLinear2Physical done, press a key\n")));

	return physAddr;
}


TUint CArloArm4::PlatExamine(TUint32 aAddressToExamine)
{
	iArlo.ExamineArm4(&aAddressToExamine);
	return aAddressToExamine;
}


/********************************
 *
 * Series 5
 *
 ********************************/

TUint32	CArloSeries5::PlatLinear2Physical(TUint32 aLinearAdr)
{
	TUint32 aPageDir = 0x40000014;

	aPageDir = ArloExamine(aPageDir);

	// Get virtual address of page directory
	DBGPRINT((_L("Page directory %08x\n"), aPageDir));
	
	TUint32 aLevelOne = aPageDir + ((aLinearAdr & 0xfff00000) >> 18);
	DBGPRINT((_L("Level One address %08x\n"), aLevelOne));
	
	aLevelOne = ArloExamine(aLevelOne);
	DBGPRINT((_L("Level One descriptor %08x\n"), aLevelOne));
	
	// we now have the level one descriptor
	if ((aLevelOne & 0x3) == 0x2) 
	{
		DBGPRINT((_L("Section\n")));
		// a Section descriptor
		return ((aLevelOne & 0xfff00000) | (aLinearAdr & 0x000fffff));
	}
	if ((aLevelOne & 0x3) != 0x1) 
	{
		DBGPRINT((_L("fault or reserved\n")));
		// Fault or reserved
		return 0;
	}

	// Page descriptor.  Get Page table
	TUint32 aLevelTwo = (aLevelOne & 0xfffffc00);
	// convert Page table to linear
	DBGPRINT((_L("physical aLevelTwo=0x%x\n"), aLevelTwo));
	aLevelTwo = RMmu::LinearAddress(aLevelTwo);
	DBGPRINT((_L("lineair aLevelTwo=0x%x\n"), aLevelTwo));
	aLevelTwo += ((aLinearAdr & 0x000ff000) >> 10);

	DBGPRINT((_L("Level Two address %08x\n"), aLevelTwo));
	aLevelTwo = ArloExamine(aLevelTwo);
	
	DBGPRINT((_L("Level Two descriptor %08x\n"), aLevelTwo));
	// level two descriptor
	if ((aLevelTwo & 0x3) == 0x2) 
	{
		// a small (4K) page
		return ((aLevelTwo & 0xfffff000) | (aLinearAdr & 0x00000fff));
	}
	if ((aLevelTwo & 0x3) == 0x1) 
	{
		// a large (64K) page
		return ((aLevelTwo & 0xffff0000) | (aLinearAdr & 0x0000ffff));
	}
	
	// fault or reserved
	return 0;
}

void CArloSeries5::PlatBoot(TUint32 aPhysMapAdr, TUint32 aBootAdr)
{
	iArlo.BootS5(aPhysMapAdr, aBootAdr);		// and we're gone
}

TUint CArloSeries5::PlatExamine(TUint32 aAddressToExamine)
{
	iArlo.ExamineS5(&aAddressToExamine);
	return aAddressToExamine;
}



/********************************
 *
 * Series 5mx
 *
 ********************************/

void CArloSeries5Mx::PlatBoot(TUint32 aPhysMapAdr, TUint32 aBootAdr)
{
	iArlo.BootS5mx(aPhysMapAdr, aBootAdr);		// and we're gone
}

void CArloSeries7::PlatBoot(TUint32 aPhysMapAdr, TUint32 aBootAdr)
{
	TBuf<64>	info;
	iArlo.BootS7(aPhysMapAdr, aBootAdr);		// and we're gone
}

/********************************
 *
 * WINS
 *
 ********************************/

void CArloWins::PlatBoot(TUint32 /*aPhysMapAdr*/, TUint32 /*aBootAdr*/)
{
	DBGPRESSKEY((_L("*** FAKE WINS KERNEL BOOT *** \n")));
}
