/*******************************************************************
 *
 *	File:		CArloConfig.cpp
 *
 *	Author:		Peter van Sebille (peter@yipton.demon.co.uk)
 *
 *	(c) Copyright 2001, Peter van Sebille
 *	All Rights Reserved
 *
 *******************************************************************/

#include <e32std.h>
#include <f32file.h>
#include <e32svr.h>
#include <e32cons.h>

#include "CArloConfig.h"

#define KKernelString		_L("image")
#define KInitRdString		_L("initrd")
#define KBootString			_L("boot")
#define KTimeoutString		_L("timeout")
#define KDefaultBoot		_L("default")
#define KMachineName		_L("machine")


CArloConfig::~CArloConfig()
{
	delete iArloBootItems;
	iArloBootItems = NULL;
}

CArloConfig::CArloConfig()
{
}

CArloConfig* CArloConfig::ReadConfigFileLC()
{
	CArloConfig*	arloConfig = new(ELeave) CArloConfig;
	CleanupStack::PushL(arloConfig);
	arloConfig->ConstructL();
	return arloConfig;
}

void CArloConfig::NewBootItem(const TDesC& aBootItem)
{
	TArloBootItem	item;

	if (iArloBootItems->Count() < 10)
	{
		item.iKernelFileName = _L("");
		item.iInitrdFileName = _L("");
		item.iDescription = aBootItem;
		item.iBootString = _L("");
		iArloBootItems->AppendL(item);
	}
}


void CArloConfig::ParseItems(const TDesC& aToken, const TDesC&  aValue)
{
	RDebug::Print(_L("  token = -%S-"), &aToken);
	RDebug::Print(_L("  value = -%S-"), &aValue);
	
	TInt	count = iArloBootItems->Count();
	if (aToken.CompareF(KKernelString) == 0)
	{
		if (count > 0 && count<10)
			iArloBootItems->At(count - 1).iKernelFileName = aValue;
	}
	else if (aToken.CompareF(KInitRdString) == 0)
	{
		if (count > 0 && count<10)
			iArloBootItems->At(count - 1).iInitrdFileName = aValue;
	}
	else if (aToken.CompareF(KBootString) == 0)
	{
		if (count > 0 && count<10)
			iArloBootItems->At(count - 1).iBootString = aValue;
	}
	else if (aToken.CompareF(KTimeoutString) == 0)
	{
		TLex	input(aValue);

		TInt	timeout;
		if (input.Val(timeout) == KErrNone)
			iTimeout = timeout;

	}
	else if (aToken.CompareF(KDefaultBoot) == 0)
	{
		iDefaultBootItemString = aValue;
	}
	else if (aToken.CompareF(KMachineName) == 0)
	{
		iMachineName = aValue;
	}
}


void CArloConfig::ParseLine(const TDesC& aLine)
{
	RDebug::Print(_L("line = %S"), &aLine);

	if (aLine.Length() != 0)
	{
		if (aLine.Left(1) != _L("#"))
		{
			if ((aLine.Left(1) == _L("[")) && aLine.Right(1) == _L("]"))
			{
				NewBootItem(aLine.Mid(1, aLine.Length() - 2));
			}
			else
			{
				TInt	pos = aLine.LocateF(TChar('='));
			
				if ((pos > 0) && (pos + 1 < aLine.Length()))
				{
					TLex		input(aLine.Left(pos));

					input.SkipSpaceAndMark();
					input.SkipCharacters();
					if (input.TokenLength() != 0)
					{
						TPtrC		token = input.MarkedToken();

						input.Assign(aLine.Mid(pos+1));
						input.SkipSpaceAndMark();
						TPtrC		value = input.Remainder();
						ParseItems(token, value);
					}
				}
			}
		}
	}
}

void CArloConfig::ReadConfigFileL(RFs& aFs, const TDesC& aFileName)
{
	RFile			file;

	User::LeaveIfError(file.Open(aFs, aFileName, EFileRead));

	TInt size;
	User::LeaveIfError(file.Size(size));

	HBufC *buf = HBufC::NewLC(size);
	TPtr	des = buf->Des();
	User::LeaveIfError(file.Read(des, size));

	TBool	finished = EFalse;
	while(!finished)
	{
		TInt pos = des.LocateF(TChar('\n'));
		if (pos >= 0)
		{
			ParseLine(des.Left(pos - 1));
			des = des.Mid(pos + 1);
		}
		else
		{
			ParseLine(des);
			finished = ETrue;
		}
	}

	file.Close();
	CleanupStack::PopAndDestroy();		// buf

	iConfigFile = aFileName;
	if ((iTimeout < -1) || (iTimeout > 30))
	{
		iTimeout = 30;
	}
}


void CArloConfig::Defaults()
{
	iTimeout = -1;
	iDefaultBootItem = 0;
	iArloBootItems->Reset();
	iDefaultBootItemString = _L("");
	iConfigFile = _L("<no config file>");

	// set default machine name
	TMachineInfoV1Buf		machineInfo;
	TInt err = UserHal::MachineInfo(machineInfo);
	if (err == KErrNone)
	{
		iMachineName = machineInfo().iMachineName;
	}
	else
	{
		iMachineName = _L("<unknown machine>");
	}
}


void CArloConfig::SetDefaultBootItem()
{
	TInt count = iArloBootItems->Count();
	for (TInt i=0; i<count ; i++)
	{
		if (iArloBootItems->At(i).iDescription.CompareF(iDefaultBootItemString) == 0)
		{
			iDefaultBootItem = i;
			break;
		}
	}
}


void CArloConfig::ConstructL()
{
	iArloBootItems = new(ELeave) CArrayFixFlat<TArloBootItem>(5);

	RFs		fs;

	User::LeaveIfError(fs.Connect());

	TInt	err;
	Defaults();
	TRAP(err, ReadConfigFileL(fs, _L("c:\\arlo.cfg")));
	if (err)
	{
		Defaults();
		TRAP(err, ReadConfigFileL(fs, _L("d:\\arlo.cfg")));
		if (err)
		{
			Defaults();
		}
	}
	if (err == KErrNone)
	{
		SetDefaultBootItem();
	}
	fs.Close();
}
